package org.openntf.wf.xsp.beans;

import java.io.Serializable;
import java.util.List;

import javax.faces.FacesException;
import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.context.FacesContext;

import org.openntf.wf.kernel.IWorkflowNotificationProvider;
import org.openntf.wf.kernel.IWorkflowProtocolStore;
import org.openntf.wf.kernel.IWorkflowStep;
import org.openntf.wf.kernel.IWorkflowStepProvider;
import org.openntf.wf.kernel.IWorkflowVariableProvider;
import org.openntf.wf.kernel.WorkflowObject;
import org.openntf.wf.kernel.WorkflowProcessor;
import org.openntf.wf.kernel.WorkflowProtocolEntry;
import org.openntf.wf.kernel.WorkflowVariable;
import org.openntf.wf.kernel.WorkfowValidationException;
import org.openntf.wf.kernel.rules.IRuleEvaluator;
import org.openntf.wf.kernel.rules.RuleEvaluatorProcessor;
import org.openntf.wf.xsp.providers.ELRuleEvaluator;

import com.ibm.commons.util.StringUtil;
import com.ibm.xsp.FacesExceptionEx;
import com.ibm.xsp.application.ApplicationEx;
import com.ibm.xsp.extlib.util.ExtLibUtil;
import com.ibm.xsp.util.FacesUtil;

import lotus.domino.NotesException;

/**
 * The WorkflowBean acts as a facade to the XSP Implementation of the OpenNTF Workflow Maschine.
 * All calls should go thru this API.<br/>
 * 
 * The WorklfowBean expect in the META-INF/services following files with a reference to a on class of a defined Type:<br/><br/>
 * <table>
 * <tr><td>Filename</td><td>Interface</td><td>Abstract implementation</td></tr>
 * 
 * <tr><td>org.openntf.wf.xsp.protocolstorageprovider</td><td>@see IWorkflowProtocolStore</td><td>@see AbstractNotesProtocolStorageProvider</td></tr>
 * <tr><td>org.openntf.wf.xsp.stepstorageprovider</td><td>@see IWorkflowStepProvider</td><td>@see AbstractNotesStepStorageProvider</td></tr>
 * <tr><td>org.openntf.wf.xsp.variablestorageprovider</td><td>@see IWorkflowVariableProvider</td><td>@see AbstractNotesVariableStoargeProvider</td></tr>
 * <tr><td>org.openntf.wf.xsp.notificationstorageprovider</td><td>@see IWorkflowNotificationProvider</td><td>@see AbstractNotesNotificationStorageProvider</td></tr>
 * </table>
 * 
 * @author Christian Guedemann
 */
/**
 * @author Administrator
 *
 */
public class WorkflowBean implements Serializable {

	private static final long serialVersionUID = 1L;
	public static final String BEAN_NAME = "wfBean";
	private static String PROTOCOLSTORAGE = "org.openntf.wf.xsp.protocolstorageprovider";
	private static String STEPSTORAGE = "org.openntf.wf.xsp.stepstorageprovider";
	private static String VARIABLESTORAGE = "org.openntf.wf.xsp.variablestorageprovider";
	private static String NOTIFICATIONSTORAGE = "org.openntf.wf.xsp.notificationstorageprovider";

	private String applicationId;
	private WorkflowProcessor processor;

	private String applicationTitle;
	private boolean isRunning;

	public WorkflowBean() {
		processApplicationInfos();
		initProcessor();
	}

	/**
	 * Getts the WorkflowBean for the current FacesContext.
	 * 
	 * @return WorkflowBean
	 */
	public static WorkflowBean get() {
		return get(FacesContext.getCurrentInstance());
	}

	/**
	 * Getting the WorkflowBean from a defined conext
	 * 
	 * @param context
	 * @return WorkflowBean
	 */
	public static WorkflowBean get(FacesContext context) {
		return (WorkflowBean) context.getApplication().getVariableResolver().resolveVariable(context, BEAN_NAME);
	}

	/**
	 * Short hand to move a business object to the next workflow step or to
	 * initiate the workflow.
	 * 
	 * @param workflowId
	 *            The ID of the workflow
	 * @param boId
	 *            The unique id of the business object
	 * @param businessObject
	 *            The business object it self's
	 * @return true if it was sucessfull, otherwise error are posted as
	 *         FaceMessages into the current context
	 */
	public boolean proceed(String workflowId, String boId, Serializable businessObject) {
		WFProceedExecutor proceedExecutor = new WFProceedExecutor(this, workflowId, boId, businessObject);
		return proceedExecutor.execute();
	}

	/**
	 * Builds a WFProceedExecutor and allows you to customize the Executor. The
	 * execution is invoked with "execute"
	 * 
	 * @param workflowId
	 *            The id of the workflow
	 * @param boId
	 *            The unique id of the business object
	 * @param businessObject
	 *            The business object
	 * @return A instance of @see WFProceedExecutor
	 */
	public WFProceedExecutor buildProceedExecutor(String workflowId, String boId, Serializable businessObject) {
		return new WFProceedExecutor(this, workflowId, boId, businessObject);
	}

	@SuppressWarnings({ "unchecked" })
	private void initProcessor() {
		List<IWorkflowProtocolStore> protocolStores = ApplicationEx.getInstance().findServices(PROTOCOLSTORAGE);
		List<IWorkflowStepProvider> stepStores = ApplicationEx.getInstance().findServices(STEPSTORAGE);
		List<IWorkflowVariableProvider> varStores = ApplicationEx.getInstance().findServices(VARIABLESTORAGE);
		List<IWorkflowNotificationProvider> notificationStores = ApplicationEx.getInstance().findServices(NOTIFICATIONSTORAGE);
		if (protocolStores.isEmpty() || stepStores.isEmpty() || varStores.isEmpty() || notificationStores.isEmpty()) {
			return;
		}
		IWorkflowStepProvider stepProvider = stepStores.get(0);
		processor = new WorkflowProcessor(protocolStores.get(0), stepProvider, getRuleEvaluatorImpl(), notificationStores.get(0), varStores.get(0));
		isRunning = true;
	}

	private void checkIsRunning() {
		if (!isRunning) {
			throw new FacesException("WF Engine not initialized! Please check configuration.");
		}
	}

	private IRuleEvaluator getRuleEvaluatorImpl() {
		return new ELRuleEvaluator();
	}

	/**
	 * Checks if a business object is in a particular workflow (based on the id
	 * of the business object and the id of the workflow)
	 * 
	 * @param boId
	 *            The unique id of the business object
	 * @param workflowId
	 *            The unique id of the workflow
	 * @return true if ther is an active workflow
	 */
	public boolean isBusinessObjectInWorkflow(String boId, String workflowId) {
		checkIsRunning();
		return processor.isBusinessObjectInWorkflow(getApplicationId(), boId, workflowId);
	}

	/**
	 * Checks if I (my user) is allowed to finish the current step
	 * 
	 * @param wfObject
	 *            The workflow representation of a particular business object
	 * @return true if I'm responsible in the current step
	 */
	public boolean canICompleteThisStep(WorkflowObject wfObject) {
		checkIsRunning();
		try {
			return processor.canICompleteThisStep(wfObject, getEffectiveUserNameOfCurrentUser());
		} catch (Exception ex) {
			throw new FacesException(ex);
		}

	}

	private String getEffectiveUserNameOfCurrentUser() throws NotesException {
		return ExtLibUtil.getCurrentSession().getEffectiveUserName();
	}

	/**
	 * Creates a new WorkflowObject based on a business object, its unique id
	 * and the workflow id
	 * 
	 * @param boId
	 *            The unique id of the business object
	 * @param businessObject
	 *            The business object
	 * @param wfId
	 *            The id of the workflow
	 * @return A new WorkflowObject (which should be saved)
	 */
	public WorkflowObject createWorkflowObjectFromBusinessObject(String boId, Serializable businessObject, String wfId) {
		checkIsRunning();
		WorkflowProcessor.WorkflowObjectBuilder builder = WorkflowProcessor.getBuilder().boId(boId).businessObject(businessObject).wfId(wfId);
		try {
			builder.appId(getApplicationId());
			builder.applicationName(getApplicationTitle());
		} catch (Exception ex) {
			throw new FacesException("Error getting Database Informations", ex);
		}
		try {
			return processor.createWorkflowObjectFromBusinessObject(builder);
		} catch (Exception ex) {
			throw new FacesException(ex);
		}

	}

	/**
	 * Accesses the current WorkflowStep of this WorkflowObject
	 * 
	 * @param wfObject
	 *            The workflow representation of a particular business object
	 * @return The current WorkflowStep
	 */
	public IWorkflowStep getWorkflowStep(WorkflowObject wfObject) {
		checkIsRunning();
		return processor.getWorkflowStep(wfObject);
	}

	public IWorkflowStep getWorkflowStepById(String wfId, String stepId) {
		return processor.getWfStepProvider().getStepByID(wfId, stepId);
	}

	/**
	 * Loads the WorklfowObject for a business object based on its unique id and
	 * the id of the workflow
	 * 
	 * @param boId
	 *            The unique id of the business object
	 * @param businessObject
	 *            The business Object
	 * @param workflowId
	 *            The id of the worklfow
	 * @return The workflow representation of this business object in the
	 *         specified workflow
	 */
	public WorkflowObject loadWorkflowObjectFromBusinessObject(String boId, Serializable businessObject, String workflowId) {
		checkIsRunning();
		return processor.loadWorkflowObjectFromBusinessObject(getApplicationId(), boId, workflowId, businessObject);
	}

	/**
	 * Process the WorkflowObject to the next step. There is no check, if the
	 * user is allowed to do, and also no validation executed. This phases have
	 * separate methods
	 * 
	 * @param wfObject
	 *            The WorkflowObject
	 */
	public void process2NextStep(WorkflowObject wfObject) {
		checkIsRunning();
		try {
			processor.process2NextStep(wfObject, getEffectiveUserNameOfCurrentUser());
		} catch (Exception ex) {
			throw new FacesException(ex);
		}
	}

	/**
	 * Get all the ProtocolEntries for a business object in a specific workflow
	 * 
	 * @param boId
	 *            The unique id of the business object
	 * @param workflowId
	 *            The id of the workflow
	 * @return A unsorted List of WorkflowProtocolEntries
	 */
	public List<WorkflowProtocolEntry> getAllProtocolEntries(String boId, String workflowId) {
		checkIsRunning();
		return processor.getAllProtocolEntries(getApplicationId(), boId, workflowId);
	}

	/**
	 * Save a WorkflowObject to the WorkflowProtocolProvider
	 * 
	 * @param wfObject
	 *            The WorklfowObject
	 */
	public void saveWorkflowObject(WorkflowObject wfObject) {
		checkIsRunning();
		processor.saveWorkflowObject(wfObject);
	}

	/**
	 * Validates the WorkflowObject. The validation messages are direct assigned
	 * as FacesMessages in the current context
	 * 
	 * @param wfObject
	 *            The WorkflowObject
	 * @return true if the object meets all WorkflowValidationRules
	 */
	public boolean validateThisStep(WorkflowObject wfObject) {
		return validateThisStep(wfObject, null);
	}

	/**
	 * Validates the WorkflowObject. The validation messages are direct assigned
	 * as FacesMessages in the current context
	 * 
	 * @param wfObject
	 *            The WorkflowObject
	 * @param targetComponent
	 *            The name of the UIComponent, where the messages should be
	 *            assigned to
	 * @return true if the object meets all WorkflowValidationRules
	 */
	public boolean validateThisStep(WorkflowObject wfObject, String targetComponent) {
		checkIsRunning();
		try {
			processor.validateThisStep(wfObject);
		} catch (WorkfowValidationException ex) {
			processMessagesAsError(targetComponent, ex.getValidationMessages().toArray(new String[0]));
			return false;
		}
		return true;
	}

	/**
	 * Returns a list of global variables
	 * 
	 * @return List of global defined WorkflowVariable
	 */
	public List<WorkflowVariable> getGlobalVariables() {
		checkIsRunning();
		return processor.getGlobalVariables();
	}

	/**
	 * The current id of the application
	 * 
	 * @return The replication id
	 */
	public String getApplicationId() {
		checkIsRunning();
		return applicationId;
	}

	/**
	 * The current title of the application
	 * 
	 * @return The title of the application
	 */
	public String getApplicationTitle() {
		checkIsRunning();
		return applicationTitle;
	}

	private synchronized void processApplicationInfos() {
		try {
			applicationId = ExtLibUtil.getCurrentDatabase().getReplicaID();
			applicationTitle = ExtLibUtil.getCurrentDatabase().getTitle();
		} catch (NotesException ex) {
			throw new FacesExceptionEx("Error during evaluation of Application Infos", ex);
		}
	}

	private void processMessagesAsError(String targetComponente, String... messages) {
		FacesContext ctx = FacesContext.getCurrentInstance();
		UIComponent c = null;
		if (!StringUtil.isEmpty(targetComponente)) {
			UIComponent start = ctx.getViewRoot();
			c = FacesUtil.getComponentFor(start, targetComponente);

		}
		for (String msg : messages) {
			FacesMessage m = new FacesMessage(FacesMessage.SEVERITY_ERROR, msg, msg);
			ctx.addMessage(c != null ? c.getClientId(ctx) : null, m);
		}
	}

	/**
	 * Check if a wfObject is in the last step of a workflow.
	 * 
	 * @param wfObject
	 *            The Workflow Object
	 * @return true if the assigned step is the last step
	 */
	public boolean isLastStep(WorkflowObject wfObject) {
		return processor.isLastStep(wfObject);
	}

	/**
	 * Finishes the wfObject. Regular indicates if the finish was triggered by a
	 * wfStep that isLastStep()
	 * 
	 * @param wfObject
	 *            The Workflow Object
	 * @param regularEnd
	 *            Indicator if this finish was regular, based on a WorkflowStep
	 *            that is a last step
	 */
	public void finishWofklow(WorkflowObject wfObject, boolean regularEnd) {
		checkIsRunning();
		try {
			processor.finishWorkflow(wfObject, getEffectiveUserNameOfCurrentUser(), regularEnd);
		} catch (Exception ex) {
			throw new FacesException(ex);
		}
	}

	protected boolean executeProceed(WFProceedExecutor proceedExecutor) {
		if (isBusinessObjectInWorkflow(proceedExecutor.getBoId(), proceedExecutor.getWorkflowId())) {
			WorkflowObject wfObject = loadWorkflowObjectFromBusinessObject(proceedExecutor.getBoId(), proceedExecutor.getBusinessObject(), proceedExecutor.getWorkflowId());
			if (!canICompleteThisStep(wfObject)) {
				processMessagesAsError(proceedExecutor.getTargetComponent(), proceedExecutor.getNotResponsibleMessage());
				return false;
			}
			if (!validateThisStep(wfObject, proceedExecutor.getTargetComponent())) {
				return false;
			}
			IWorkflowStep step = getWorkflowStep(wfObject);
			if (step.isLastStep()) {
				finishWofklow(wfObject, true);
			} else {
				process2NextStep(wfObject);
			}
			return true;
		} else {
			WorkflowObject wfObject = createWorkflowObjectFromBusinessObject(proceedExecutor.getBoId(), proceedExecutor.getBusinessObject(), proceedExecutor.getWorkflowId());
			saveWorkflowObject(wfObject);
			return true;
		}

	}

	public void enforce2WorkflowStep(WorkflowObject wfObject, String stepId) {
		checkIsRunning();
		try {
			processor.enforce2WorkflowStep(wfObject, stepId, getEffectiveUserNameOfCurrentUser());
		} catch (Exception ex) {
			throw new FacesException(ex);
		}
	}

	public IWorkflowNotificationProvider getWfNotficationProvider() {
		return processor.getWfNotficationProvider();
	}

	public IWorkflowProtocolStore getWfProtocolStore() {
		return processor.getWfProtocolStore();
	}

	public RuleEvaluatorProcessor getWfRuleEvaluatorProcessor() {
		return processor.getWfRuleEvaluatorProcessor();
	}

	public IWorkflowStepProvider getWfStepProvider() {
		return processor.getWfStepProvider();
	}

	public IWorkflowVariableProvider getWfVariableProvider() {
		return processor.getWfVariableProvider();
	}
}
