package org.openntf.wf.xsp.providers;

import java.util.ArrayList;
import java.util.List;

import org.openntf.wf.kernel.IWorkflowHeader;
import org.openntf.wf.kernel.IWorkflowStep;
import org.openntf.wf.kernel.IWorkflowStepProvider;
import org.openntf.wf.kernel.WorkflowRuntimeException;
import org.openntf.wf.kernel.impl.WorkflowStepImpl;
import org.openntf.wf.kernel.impl.WorkflowHeaderImpl;
import org.openntf.wf.kernel.rules.RuleEvaluatorProcessor;
import org.openntf.wf.kernel.rules.WorkflowRoutingRule;
import org.openntf.wf.kernel.rules.WorkflowValidationRule;
import org.openntf.wf.xsp.util.Utils;

import lotus.domino.Database;
import lotus.domino.Document;
import lotus.domino.DocumentCollection;
import lotus.domino.NotesException;
import lotus.domino.View;

public abstract class AbstractNotesStepStorageProvider implements IWorkflowStepProvider {

	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private RuleEvaluatorProcessor processor;

	@Override
	public IWorkflowStep getFirstStep(String wfId) {
		NotesAccessProperties accessProperties = null;
		try {
			accessProperties = accessFirstStepDocument(wfId);
			return buildWFStep(wfId, accessProperties.getWfDocument().getItemValueString("ID"), accessProperties, true);

		} catch (NotesException ex) {
			throw new WorkflowRuntimeException("General Error during get first WF Step (" + wfId + ")", ex);
		} finally {
			if (accessProperties != null) {
				Utils.recyler(accessProperties.getAll4Recycle());
			}

		}
	}

	private NotesAccessProperties accessFirstStepDocument(String wfId) throws NotesException {
		NotesAccessProperties accessProperties;
		accessProperties = new NotesAccessProperties(getStepStoreDatabase());
		if (accessProperties.getWfDatabase() == null) {
			throw new WorkflowRuntimeException("No Stepstore defined", null);
		}
		accessProperties.accessView(getFirstStepViewName());
		accessProperties.accessDocument(wfId);
		return accessProperties;
	}

	private String getFirstStepViewName() {
		return "LUPWFStepFirstByWFHeaderID";
	}

	@Override
	public IWorkflowStep getStepByID(String wfId, String stepId) {
		return intGetStepById(wfId, stepId, true);
	}

	private IWorkflowStep intGetStepById(String wfId, String stepId, boolean full) {
		NotesAccessProperties accessProperties = null;
		try {
			accessProperties = accessStepDocument(wfId + "@@@" + stepId);
			return buildWFStep(wfId, stepId, accessProperties, full);

		} catch (NotesException ex) {
			throw new WorkflowRuntimeException("General Error during load (" + wfId + " / " + stepId + ")", ex);
		} finally {
			if (accessProperties != null) {
				if (full) {
					Utils.recyler(accessProperties.getAll4Recycle());
				} else {
					Utils.recyler(accessProperties.getWfDocument());
				}
			}

		}
	}

	@SuppressWarnings("unchecked")
	private IWorkflowStep buildWFStep(String wfId, String stepId, NotesAccessProperties accessProperties, boolean full) throws NotesException {
		WorkflowStepImpl wfStep = new WorkflowStepImpl(processor);
		wfStep.setStepId(stepId);
		wfStep.setStepName(accessProperties.getWfDocument().getItemValueString("title"));
		wfStep.setWorkflowId(wfId);
		if (full) {
			wfStep.setNextStepRules(getWorkflowRoutingRules(wfId, stepId, accessProperties));
			wfStep.setValidationRules(getValidationRules(wfId, stepId, accessProperties));
		}
		wfStep.setResponsibles(accessProperties.getWfDocument().getItemValue("workflowresponsibles"));
		wfStep.setResponsibleCalculated("1".equals(accessProperties.getItemValueString("responsiblesCalculated")));
		wfStep.setResponsibleExpression(accessProperties.getItemValueString("responsiblesExpression"));
		wfStep.setEnteringStepOperation(accessProperties.getItemValueString("enteringOperation"));
		wfStep.setPreStepOperation(accessProperties.getItemValueString("preStepOperation"));
		wfStep.setPostStepOperation(accessProperties.getItemValueString("postStepOperation"));
		return wfStep;
	}

	private List<WorkflowValidationRule> getValidationRules(String wfId, String stepId, NotesAccessProperties accessProperties) throws NotesException {
		View lupValidationRules = accessProperties.getWfDatabase().getView(getValidationRuleViewName());
		DocumentCollection rulesCollection = lupValidationRules.getAllDocumentsByKey(stepId, true);
		List<WorkflowValidationRule> rules = new ArrayList<WorkflowValidationRule>();
		Document docNext = rulesCollection.getFirstDocument();
		while (docNext != null) {
			Document docProcess = docNext;
			docNext = rulesCollection.getNextDocument();

			WorkflowValidationRule rule = new WorkflowValidationRule();
			rule.setExpression(docProcess.getItemValueString("expression"));
			rule.setMessage(docProcess.getItemValueString("message"));
			rule.setRuleId(docProcess.getItemValueString("ID"));
			rule.setTestOrder(docProcess.getItemValueInteger("testOrder"));
			rules.add(rule);
			docProcess.recycle();
		}
		Utils.recyler(lupValidationRules, rulesCollection);
		return rules;
	}

	private String getValidationRuleViewName() {
		return "lupWFValidationRuleByStepID";
	}

	private List<WorkflowRoutingRule> getWorkflowRoutingRules(String wfId, String stepId, NotesAccessProperties accessProperties) throws NotesException {
		View lupStepRules = accessProperties.getWfDatabase().getView(getNextStepRuleViewName());
		DocumentCollection rulesCollection = lupStepRules.getAllDocumentsByKey(stepId, true);
		List<WorkflowRoutingRule> rules = new ArrayList<WorkflowRoutingRule>();
		Document docNext = rulesCollection.getFirstDocument();
		while (docNext != null) {
			Document docProcess = docNext;
			docNext = rulesCollection.getNextDocument();

			WorkflowRoutingRule rule = new WorkflowRoutingRule();
			rule.setExpression(docProcess.getItemValueString("expression"));
			rule.setRuleId(docProcess.getItemValueString("ID"));
			rule.setTestOrder(docProcess.getItemValueInteger("testOrder"));
			rule.setNextStep(intGetStepById(wfId, docProcess.getItemValueString("nextStepId"), false));
			rules.add(rule);
			docProcess.recycle();
		}
		Utils.recyler(lupStepRules, rulesCollection);
		return rules;
	}

	private String getNextStepRuleViewName() {
		return "lupWFStepRuleByWFStepID";
	}

	@Override
	public void setRuleEvaluationProcessor(RuleEvaluatorProcessor processor) {
		this.processor = processor;
	}

	protected abstract Database getStepStoreDatabase() throws NotesException;

	private NotesAccessProperties accessStepDocument(String accessKey) throws NotesException {
		NotesAccessProperties accessProperties;
		accessProperties = new NotesAccessProperties(getStepStoreDatabase());
		if (accessProperties.getWfDatabase() == null) {
			throw new WorkflowRuntimeException("No Stepstore defined", null);
		}
		accessProperties.accessView(getStepViewName());
		accessProperties.accessDocument(accessKey);
		return accessProperties;
	}

	private String getStepViewName() {
		return "lupWFStepByWFHeaderAndWFStepID";
	}

	@Override
	public IWorkflowHeader getHeader(String workflowId) {
		NotesAccessProperties nap = null;
		try {
			nap = new NotesAccessProperties(getStepStoreDatabase());
			nap.accessView("lupWFHeaderByID");
			nap.accessDocument(workflowId);
			return buildWFHeader(nap, workflowId);
		} catch (Exception e) {
			throw new WorkflowRuntimeException("No Header found!", e);
		} finally {
			Utils.recyler(nap.getAll4Recycle());

		}
	}

	private IWorkflowHeader buildWFHeader(NotesAccessProperties nap, String workflowId) throws NotesException {
		String name = nap.getItemValueString("workflowTitle");
		String comment = nap.getItemValueString("comment");
		@SuppressWarnings("unchecked")
		List<String> lstOwner = nap.getWfDocument().getItemValue("workflowOwner");
		return new WorkflowHeaderImpl(workflowId, name, lstOwner, comment);
	}
}
